import os
import hashlib
import requests
from urllib.parse import urlparse
import shutil
import appdirs
import pandas as pd


def get_data_file(relative_path):
    script_dir = os.path.dirname(
        __file__
    )  # get the directory of the current script
    data_file_path = os.path.join(
        script_dir, "data", relative_path
    )  # build the path to the data file
    return data_file_path


def get_test_data_file(relative_path):
    script_dir = os.path.dirname(
        __file__
    )  # get the directory of the current script
    data_file_path = os.path.join(
        script_dir, "test/data", relative_path
    )  # build the path to the data file
    return data_file_path


def load_test_data_file(relative_path):
    test_sample = pd.read_csv(get_test_data_file(relative_path), index_col=0)
    return test_sample


def cached_download(url_or_filepath):
    """Downloads the file at a URL and saves it locally. If called again with the same URL it will use the saved file. Returns the local filepath"""
    # Hash the URL to create a unique filename
    if os.path.isfile(url_or_filepath):
        # If the provided URL is a local file path, return it directly
        return url_or_filepath

    url = url_or_filepath
    url_path = urlparse(url).path
    ext = os.path.splitext(url_path)[1]
    filename = hashlib.sha256(url.encode()).hexdigest() + ext

    app_name = "bio-learn"
    download_path = appdirs.user_cache_dir(app_name)

    # Ensure download path exists
    os.makedirs(download_path, exist_ok=True)

    filepath = os.path.join(download_path, filename)

    if os.path.exists(filepath):
        # If the file is already downloaded, return the file path
        return filepath
    else:
        # Try to download the file
        response = requests.get(url, stream=True)
        response.raise_for_status()  # Raise an HTTPError if one occurred

        # If the file is not downloaded yet, download and save it
        with open(filepath, "wb") as out_file:
            shutil.copyfileobj(response.raw, out_file)

        # Return the file path
        return filepath



import pandas as pd
from biolearn.model import (
    model_definitions,
    LinearMethylationModel,
    GrimageModel,
    SexEstimationModel,
    ImputationDecorator,
)
from biolearn.imputation import (
    biolearn_impute,
    hybrid_impute,
    impute_from_average,
)
from biolearn.util import get_data_file


class ModelGallery:
    """
    ModelGallery manages a collection of Models that can be run on biological data to produce predictions.
    It supports retrieving models by name with various imputation methods, and searching models based on species or tissue.
    """

    model_builders = {
        "LinearMethylationModel": LinearMethylationModel.from_definition,
        "GrimageModel": GrimageModel.from_definition,
        "SexEstimationModel": SexEstimationModel.from_definition,
    }

    def __init__(self, models=model_definitions):
        """
        Initializes the ModelGallery instance.

        Args:
            models (dict): A dictionary of model definitions.
        """
        self.models = {}
        for name, model_def in models.items():
            if not isinstance(model_def, dict):
                raise ValueError(
                    f"Expected dictionary for model definition, got {type(model_def)} for model {name}"
                )
            model_type = model_def["model"]["type"]
            if model_type in self.model_builders:
                self.models[name] = self.model_builders[model_type](model_def)
            else:
                raise ValueError(
                    f"Model type {model_type} does not have a known builder"
                )

    def get(self, name, imputation_method="default"):
        """
        Retrieves a model by its name with specified imputation method. Possible imputation values are:
        'none', 'biolearn', 'averaging', 'dunedin', 'sesame_450k'. If no value is passed it will default to sesame_450k.

        Args:
            name (str): The name of the model.
            imputation_method (str, optional): The imputation method to use

        Returns:
            object: The requested model instance, possibly wrapped in an ImputationDecorator.

        Raises:
            KeyError: If the model with the specified name is not found.
            ValueError: If an invalid imputation method is specified.
        """
        if name not in self.models:
            raise KeyError(f"Model not found: {name}")

        # Selecting imputation method
        model_instance = self.models[name]
        if imputation_method == "none":
            return model_instance
        elif imputation_method == "biolearn":
            return ImputationDecorator(
                model_instance, lambda dnam, cpgs: biolearn_impute(dnam)
            )
        elif imputation_method == "averaging":
            return ImputationDecorator(
                model_instance,
                lambda dnam, cpgs: impute_from_average(dnam, cpgs),
            )
        elif imputation_method == "dunedin":
            # DunedinPACE method
            gold_means_file = get_data_file("DunedinPACE_Gold_Means.csv")
            df = pd.read_csv(gold_means_file, index_col=0)
            gold_averages = df["mean"]
            return ImputationDecorator(
                model_instance,
                lambda dnam, cpgs: hybrid_impute(dnam, gold_averages, cpgs),
            )
        elif imputation_method in ("sesame_450k", "default"):
            # Sesame method for 450k arrays
            sesame_450k_file = get_data_file("sesame_450k_median.csv")
            df = pd.read_csv(sesame_450k_file, index_col=0)
            gold_medians = df["median"]
            return ImputationDecorator(
                model_instance,
                lambda dnam, cpgs: hybrid_impute(dnam, gold_medians, cpgs),
            )
        else:
            raise ValueError(f"Invalid imputation method: {imputation_method}")

    def search(self, species=None, tissue=None):
        """
        Search models based on species and tissue criteria.

        Args:
            species (str, optional): The species to filter models by.

            tissue (str, optional): The tissue to filter models by.

        Returns:
            dict: A dictionary of models that match the specified criteria.
        """
        matches = {}
        for name, model in self.models.items():
            if (species is None or model.metadata["species"] == species) and (
                tissue is None or model.metadata["tissue"] == tissue
            ):
                matches[name] = model
        return matches



import yaml
import pandas as pd
import re
from biolearn.util import cached_download, get_data_file


def parse_after_colon(s):
    """Extract and return the substring after the first colon."""

    if isinstance(s, str) and s.strip() and ":" in s:
        return s.split(":")[1].strip()

    return ""


def sex_parser(s):
    if isinstance(s, str):
        s_lower = s.lower().strip()
        if s_lower in ["female", "f"]:
            return 1
        elif s_lower in ["male", "m"]:
            return 2
    return 0


def extract_numeric(s):
    """Extract the first numeric value from a string."""
    match = re.search(r"(\d+(\.\d+)?)", s)
    return float(match.group(1)) if match else None


class GeoData:
    """
    Represents genomic data with a focus on metadata and methylation data.

    GeoData facilitates the organization and access to metadata and methylation data.

    Attributes:
        metadata (DataFrame): A pandas DataFrame where rows represent different samples
                              and columns represent different data fields.
        dnam (DataFrame): A pandas DataFrame where columns represent different samples
                          and rows represent different methylation sites.
    """

    def __init__(self, metadata, dnam):
        """
        Initializes the GeoData instance.

        Args:
            metadata (DataFrame): Metadata associated with genomic samples.
            dnam (DataFrame): Methylation data associated with genomic samples.
        """
        self.metadata = metadata
        self.dnam = dnam

    def copy(self):
        """
        Creates a deep copy of the GeoData instance.

        Returns:
            GeoData: A new instance of GeoData with copies of the metadata and dnam DataFrames.
        """
        return GeoData(
            self.metadata.copy(deep=True), self.dnam.copy(deep=True)
        )


class GeoMatrixParser:
    parsers = {
        "numeric": lambda s: extract_numeric(parse_after_colon(s)),
        "string": lambda s: parse_after_colon(s),
        "sex": lambda s: sex_parser(parse_after_colon(s)),
    }

    def __init__(self, data):
        if data.get("id-row") is None:
            raise ValueError("Parser not valid: missing id-row")
        self.id_row = data.get("id-row")
        self.metadata = data.get("metadata")
        self.matrix_start = data.get("matrix-start")

    def parse(self, file_path):
        load_list = self._metadata_load_list()
        load_rows = [x[1] for x in load_list]
        column_names = [x[0] for x in load_list]
        metadata = pd.read_table(
            file_path,
            index_col=0,
            skiprows=lambda x: x != self.id_row - 1 and x not in load_rows,
        )
        metadata.index = column_names
        metadata = metadata.transpose()
        metadata.index.name = "id"
        for col in metadata.columns:
            parser_name = self.metadata[col]["parse"]
            parser = self.parsers[parser_name]
            metadata[col] = metadata[col].apply(parser)
        dnam = pd.read_table(
            file_path, index_col=0, skiprows=self.matrix_start - 1
        )
        dnam = dnam.drop(["!series_matrix_table_end"], axis=0)
        dnam.index.name = "id"
        return GeoData(metadata, dnam)

    def _metadata_load_list(self):
        load_list = [
            (key, self.metadata[key]["row"] - 1)
            for key in self.metadata.keys()
        ]
        load_list.sort(key=lambda x: x[1])
        return load_list


class DataSource:
    """
    Represents a single data source in the DataLibrary.

    This class encapsulates the details of a data source including metadata about the source and
    functionality to load the data.

    Raises:
        ValueError: If any of the required fields ('id', 'path', 'parser') are missing
                    during initialization.
    """

    REQUIRED_FIELDS = {
        "id": "'id' key is missing in item",
        "path": "'path' key is missing in item",
        "parser": "'parser' key is missing in item",
    }

    def __init__(self, data):
        """
        Initializes the DataSource instance. Intended to be initialized with data from a library YAML file.

        Args:
            data (dict): A dictionary containing the data source's properties.

        Raises:
            ValueError: If any of the required fields ('id', 'path', 'parser') are missing.
        """
        for field, error_message in self.REQUIRED_FIELDS.items():
            setattr(self, field, data.get(field))
            if getattr(self, field) is None:
                raise ValueError(error_message)

        self.title = data.get("title")
        self.summary = data.get("summary")
        self.format = data.get("format")
        self.organism = data.get("organism")

        self.parser = self._create_parser(data["parser"])

    def load(self):
        """
        Loads the data from the source.

        Returns:
            GeoData: An instance of the GeoData class containing the parsed geographical data.
        """
        file_path = cached_download(self.path)
        return self.parser.parse(file_path)

    def __repr__(self):
        """
        Represents the DataSource instance as a string.

        Returns:
            str: A string representation of the DataSource instance.
        """
        return f"DataSource(ID: {self.id}, Title: {self.title})"

    def _create_parser(self, parser_data):
        parser_type = parser_data.get("type")
        if parser_type is None:
            raise ValueError("Parser type is missing")
        if parser_type == "geo-matrix":
            return GeoMatrixParser(parser_data)
        # Add more parsers as needed
        raise ValueError(f"Unknown parser type: {parser_type}")


def parse_library_file(library_file):
    data = yaml.safe_load(library_file)
    if data is None:
        raise ValueError("File must be YAML format with 'items' at root")
    if "items" in data:
        data_sources = [DataSource(item) for item in data["items"]]
        return data_sources
    else:
        raise ValueError("File must be YAML format with 'items' at root")


class DataLibrary:
    """
    Manages a collection of data sources for biomarkers research.

    The DataLibrary class is responsible for loading, storing, and retrieving data sources.
    Data sources are defined in a library file and new sources can easily be added at runtime.
    Currently DNA methylation data from GEO is supported.
    """

    def __init__(self, library_file=None):
        """
        Initializes the DataLibrary instance.

        Args:
            library_file (str, optional): The file path of the library file. If None,
                                          the biolearn default library will be loaded.

        """
        self.sources = []
        if library_file is None:
            library_file = get_data_file("library.yaml")
        self.load_sources(library_file)

    def load_sources(self, library_file):
        """
        Loads data sources from a given library file appending them to the current set of data sources.

        Args:
            library_file (str): The file path of the library file to load data sources from.
        """
        with open(library_file, "r") as f:
            data_sources = parse_library_file(f)
            self.sources.extend(data_sources)

    def get(self, source_id):
        """
        Retrieves a data source by its identifier.

        Args:
            source_id (str): The identifier of the data source to retrieve.

        Returns:
            The data source with the given identifier if found, otherwise None.
        """
        for source in self.sources:
            if source.id == source_id:
                return source
        return None

    def lookup_sources(self, organism=None, format=None):
        """
        Looks up data sources based on the specified organism and/or format.

        Args:
            organism (str, optional): The organism to filter the data sources by.
            format (str, optional): The format to filter the data sources by.

        Returns:
            A list of data sources that match the specified organism and format criteria.
        """
        matches = []
        for source in self.sources:
            if (organism is None or source.organism == organism) and (
                format is None or source.format == format
            ):
                matches.append(source)
        return matches
